package org.implab.gradle.containers.cli;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ExecutionException;

import org.gradle.api.logging.Logger;

public abstract class DockerTraits {

    public final String BUILD_COMMAND = "build";
    public final String PUSH_COMMAND = "push";
    public final String RUN_COMMAND = "run";
    public final String SAVE_COMMAND = "save";
    public final String INSPECT_COMMAND = "inspect";
    public final String IMAGE_COMMAND = "image";
    public final String TAG_COMMAND = "tag";
    public final String COMPOSE_COMMAND = "compose";
    public final String UP_COMMAND = "up";
    public final String STOP_COMMAND = "stop";
    public final String RM_COMMAND = "rm";

    public abstract Logger getLogger();

    public abstract Optional<File> getWorkingDir();

    public abstract String getCliCmd();

    protected boolean checkRetCode(ProcessSpec proc, int code)
            throws InterruptedException, ExecutionException, IOException {
        if (getLogger().isInfoEnabled()) {
            proc.redirectStdout(RedirectTo.consumer(getLogger()::info))
                    .redirectStderr(RedirectTo.consumer(getLogger()::info));
        }

        getLogger().info("Starting: {}", proc.command());

        return proc.start().get() == code;
    }

    protected void exec(ProcessSpec proc) throws InterruptedException, IOException, ExecutionException {
        if (getLogger().isInfoEnabled())
            proc.redirectStdout(RedirectTo.consumer(getLogger()::info));

        if (getLogger().isErrorEnabled())
            proc.redirectStderr(RedirectTo.consumer(getLogger()::error));

        getLogger().info("Starting: {}", proc.command());

        var code = proc.start().get();
        if (code != 0) {
            getLogger().error("The process exited with code {}", code);
            throw new IOException("The process exited with error code " + code);
        }
    }

    protected ProcessSpec builder(String... args) {
        var spec = new ProcessSpec().args(getCliCmd()).args(args);

        getWorkingDir().ifPresent(spec::directory);

        return spec;
    }

    public void buildImage(String imageName, File contextDirectory, List<String> options)
            throws IOException, InterruptedException, ExecutionException {
        var spec = builder(BUILD_COMMAND)
                .args(options)
                .args("-t", imageName, contextDirectory.getAbsolutePath());

        exec(spec);
    }

    public void pushImage(String image, List<String> options)
            throws InterruptedException, IOException, ExecutionException {
        var spec = builder(PUSH_COMMAND)
                .args(options)
                .args(image);

        exec(spec);
    }

    public ProcessSpec runImage(String image, List<String> options, List<String> command)
            throws InterruptedException, IOException {
        return builder(RUN_COMMAND)
                .args(options)
                .args(image)
                .args(command);
    }

    public void stopContainer(String containerId, List<String> options) throws InterruptedException, IOException, ExecutionException {
        exec(builder(STOP_COMMAND, containerId).args(options));
    }

    public void saveImage(Set<String> images, File output)
            throws InterruptedException, IOException, ExecutionException {
        if (output.exists())
            output.delete();

        var spec = builder(SAVE_COMMAND)
                .args("-o", output.getAbsolutePath())
                .args(images);

        exec(spec);
    }

    public void tagImage(String source, String target) throws InterruptedException, IOException, ExecutionException {
        exec(builder(TAG_COMMAND, source, target));
    }

    public boolean imageExists(String imageId) throws InterruptedException, IOException, ExecutionException {
        getLogger().info("Check image {} exists", imageId);

        return checkRetCode(
                builder(IMAGE_COMMAND, INSPECT_COMMAND, "--format", "image-exists", imageId),
                0);
    }

    public boolean imageExists(File imageIdFile) {
        if (imageIdFile.exists()) {
            try {
                var imageId = Utils.readImageRef(imageIdFile);
                return imageExists(imageId.getId());
            } catch (IOException | InterruptedException | ExecutionException e) {
                getLogger().error("Failed to read imageId {}: {}", imageIdFile, e);
                return false;
            }
        }
        return false;
    }

    ProcessSpec compose(File primaryCompose, Set<String> profiles, String... extra) {
        var spec = builder(COMPOSE_COMMAND, "--file", primaryCompose.getAbsolutePath());

        for (var profile : profiles)
            spec.args("--profile", profile);

        spec.args(extra);

        return spec;
    }

    public void composeUp(File primaryCompose, Set<String> profiles) throws InterruptedException, IOException, ExecutionException {
        exec(compose(primaryCompose, profiles, UP_COMMAND, "--detach"));
    }

    public void composeStop(File primaryCompose, Set<String> profiles) throws InterruptedException, IOException, ExecutionException {
        exec(compose(primaryCompose, profiles, STOP_COMMAND));
    }

    public void composeRm(File primaryCompose, Set<String> profiles, boolean removeVolumes)
            throws InterruptedException, IOException, ExecutionException {
        var spec = compose(primaryCompose, profiles, RM_COMMAND, "--force", "--stop");
        if (removeVolumes)
            spec.args("--volumes");

        exec(spec);
    }
}
